// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <atomic>
#include <stdio.h>
#include <string>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <vector>

#include "source.h"
#include "filter.h"
#include "cfg.h"
#include "utils.h"
#include "log.h"
#include "view.h"
#include "picio.h"
#include "error.h"

void send_mjpeg_stream(int cfd, source *s, double fps, int quality, bool get, int time_limit, const std::vector<filter *> *const filters, std::atomic_bool *const global_stopflag, resize *const r, const int resize_w, const int resize_h, configuration_t *const cfg, const bool is_view_proxy, const bool repeat)
{
        constexpr char reply_headers[] =
                "HTTP/1.0 200 OK\r\n"
                "Cache-Control: no-cache\r\n"
                "Pragma: no-cache\r\n"
		"Server: " NAME " " VERSION "\r\n"
                "Expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
                "Connection: close\r\n"
                "Content-Type: multipart/x-mixed-replace; boundary=--myboundary\r\n"
                "\r\n";

	if (WRITE(cfd, reply_headers, strlen(reply_headers)) <= 0)
	{
		log(LL_DEBUG, "short write on response header");
		return;
	}

	if (!get)
		return;

	bool first = true;
	uint8_t *prev_frame = nullptr;
	bool sc = resize_h != -1 || resize_w != -1;
	bool nf = filters == nullptr || filters -> empty();

	uint64_t prev = 0;
	time_t end = time(nullptr) + time_limit;
	for(;(time_limit <= 0 || time(nullptr) < end) && !*global_stopflag;)
	{
		int w = -1, h = -1;
		uint8_t *work = nullptr;
		size_t work_len = 0;
		if (s -> get_frame((nf && !sc) ? E_JPEG : E_RGB, quality, &prev, &w, &h, &work, &work_len, repeat)) {
			// send header
			constexpr char term[] = "\r\n";
			if (first)
				first = false;
			else if (WRITE(cfd, term, strlen(term)) <= 0)
			{
				log(LL_DEBUG, "short write on terminating cr/lf: %s", strerror(errno));
				free(work);
				break;
			}

			// decode, encode, etc frame
			if (nf && !sc) {
				char img_h[4096] = { 0 };
				int len = snprintf(img_h, sizeof img_h, 
					"--myboundary\r\n"
					"Content-Type: image/jpeg\r\n"
					"Content-Length: %zu\r\n"
					"\r\n", work_len);
				if (WRITE(cfd, img_h, len) <= 0)
				{
					log(LL_DEBUG, "short write on boundary header: %s", strerror(errno));
					free(work);
					break;
				}

				if (WRITE(cfd, reinterpret_cast<char *>(work), work_len) <= 0)
				{
					log(LL_DEBUG, "short write on img data: %s", strerror(errno));
					free(work);
					break;
				}

				free(work);
			}
			else
			{
				source *cur_s = is_view_proxy ? ((view *)s) -> get_current_source() : s;
				instance_t *inst = find_instance_by_interface(cfg, cur_s);

				apply_filters(inst, cur_s, filters, prev_frame, work, prev, w, h);

				char *data_out = nullptr;
				size_t data_out_len = 0;
				FILE *fh = open_memstream(&data_out, &data_out_len);
				if (!fh)
					error_exit(true, "open_memstream() failed");

				bool ok = false;

				if (resize_h != -1 || resize_w != -1) {
					int target_w = resize_w != -1 ? resize_w : w;
					int target_h = resize_h != -1 ? resize_h : h;

					if (target_w != -1 && target_h != -1) {
						uint8_t *temp = nullptr;
						r -> do_resize(w, h, work, target_w, target_h, &temp);

						ok = write_JPEG_file(s -> get_meta(), fh, target_w, target_h, quality, temp);
						free(temp);
					}
				}
				else {
					ok = write_JPEG_file(s -> get_meta(), fh, w, h, quality, work);
				}

				fclose(fh);

				free(prev_frame);
				prev_frame = work;

				if (!ok) {
					log(LL_DEBUG, "write_JPEG_file failed");
					break;
				}

				char img_h[4096] = { 0 };
				int len = snprintf(img_h, sizeof img_h, 
					"--myboundary\r\n"
					"Content-Type: image/jpeg\r\n"
					"Content-Length: %zu\r\n"
					"\r\n", data_out_len);
				if (WRITE(cfd, img_h, len) <= 0)
				{
					log(LL_DEBUG, "short write on boundary header: %s", strerror(errno));
					break;
				}

				if (WRITE(cfd, data_out, data_out_len) <= 0)
				{
					log(LL_DEBUG, "short write on img data: %s", strerror(errno));
					free(data_out);
					break;
				}

				free(data_out);
				data_out = nullptr;
			}
		}

		// FIXME adapt this by how long the wait_for_frame took
		if (fps > 0) {
			double us = 1000000.0 / fps;
			if (us > 0)
				myusleep(useconds_t(us));
		}
	}

	free(prev_frame);
}

void send_mpng_stream(int cfd, source *s, double fps, bool get, const int time_limit, const std::vector<filter *> *const filters, std::atomic_bool *const global_stopflag, resize *const r, const int resize_w, const int resize_h, configuration_t *const cfg, const bool is_view_proxy, const bool repeat)
{
        constexpr char reply_headers[] =
                "HTTP/1.0 200 ok\r\n"
                "cache-control: no-cache\r\n"
                "pragma: no-cache\r\n"
		"server: " NAME " " VERSION "\r\n"
                "expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
                "connection: close\r\n"
                "content-type: multipart/x-mixed-replace; boundary=--myboundary\r\n"
                "\r\n";

	if (WRITE(cfd, reply_headers, strlen(reply_headers)) <= 0)
	{
		log(LL_DEBUG, "short write on response header");
		close(cfd);
		return;
	}

	if (!get)
	{
		close(cfd);
		return;
	}

	bool first = true;
	char *data_out = nullptr;
	size_t data_out_len = 0;

	uint8_t *prev_frame = nullptr;

	uint64_t prev = 0;
	time_t end = time(nullptr) + time_limit;
	for(;(time_limit <= 0 || time(nullptr) < end) && !*global_stopflag;)
	{
		int w = -1, h = -1;
		uint8_t *work = nullptr;
		size_t work_len = 0;
		if (s -> get_frame(E_RGB, -1, &prev, &w, &h, &work, &work_len, repeat)) {
			source *cur_s = is_view_proxy ? ((view *)s) -> get_current_source() : s;
			instance_t *inst = find_instance_by_interface(cfg, cur_s);

			apply_filters(inst, cur_s, filters, prev_frame, work, prev, w, h);

			data_out = nullptr;
			data_out_len = 0;
			FILE *fh = open_memstream(&data_out, &data_out_len);
			if (!fh)
				error_exit(true, "open_memstream() failed");

			if (resize_w != -1 || resize_h != -1) {
				int target_w = resize_w != -1 ? resize_w : w;
				int target_h = resize_h != -1 ? resize_h : h;

				uint8_t *temp = nullptr;
				r -> do_resize(w, h, work, target_w, target_h, &temp);

				write_PNG_file(fh, target_w, target_h, temp);

				free(temp);
			}
			else {
				write_PNG_file(fh, w, h, work);
			}

			fclose(fh);

			free(prev_frame);
			prev_frame = work;

			// send
			constexpr char term[] = "\r\n";
			if (first)
				first = false;
			else if (WRITE(cfd, term, strlen(term)) <= 0)
			{
				log(LL_DEBUG, "short write on terminating cr/lf");
				break;
			}

			char img_h[4096] = { 0 };
			snprintf(img_h, sizeof img_h, 
					"--myboundary\r\n"
					"Content-Type: image/png\r\n"
					"Content-Length: %d\r\n"
					"\r\n", (int)data_out_len);
			if (WRITE(cfd, img_h, strlen(img_h)) <= 0)
			{
				log(LL_DEBUG, "short write on boundary header");
				break;
			}

			if (WRITE(cfd, data_out, data_out_len) <= 0)
			{
				log(LL_DEBUG, "short write on img data");
				break;
			}

			free(data_out);
			data_out = nullptr;
		}

		// FIXME adapt this by how long the wait_for_frame took
		double us = 1000000.0 / fps;
		if (us > 0)
			myusleep(useconds_t(us));
	}

	free(data_out);

	free(prev_frame);
}

void send_png_frame(int cfd, source *s, bool get, const std::vector<filter *> *const filters, resize *const r, const int resize_w, const int resize_h, configuration_t *const cfg, const bool is_view_proxy, const bool repeat)
{
        constexpr char reply_headers[] =
                "HTTP/1.0 200 ok\r\n"
                "cache-control: no-cache\r\n"
                "pragma: no-cache\r\n"
		"server: " NAME " " VERSION "\r\n"
                "expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
                "connection: close\r\n"
                "content-type: image/png\r\n"
                "\r\n";

	if (WRITE(cfd, reply_headers, strlen(reply_headers)) <= 0) {
		log(LL_DEBUG, "short write on response header");
		close(cfd);
		return;
	}

	if (!get) {
		close(cfd);
		return;
	}

	set_no_delay(cfd, true);

	uint64_t prev_ts = get_us();
	int w = -1, h = -1;
	uint8_t *work = nullptr;
	size_t work_len = 0;
	if (s -> get_frame(E_RGB, -1, &prev_ts, &w, &h, &work, &work_len, repeat)) {
		char *data_out = nullptr;
		size_t data_out_len = 0;

		FILE *fh = open_memstream(&data_out, &data_out_len);
		if (!fh)
			error_exit(true, "open_memstream() failed");

		source *cur_s = is_view_proxy ? ((view *)s) -> get_current_source() : s;
		instance_t *inst = find_instance_by_interface(cfg, cur_s);

		apply_filters(inst, cur_s, filters, nullptr, work, prev_ts, w, h);

		if (resize_h != -1 || resize_w != -1) {
			int target_w = resize_w != -1 ? resize_w : w;
			int target_h = resize_h != -1 ? resize_h : h;

			uint8_t *temp = nullptr;
			r -> do_resize(w, h, work, target_w, target_h, &temp);

			write_PNG_file(fh, target_w, target_h, temp);

			free(temp);
		}
		else {
			write_PNG_file(fh, w, h, work);
		}

		free(work);

		fclose(fh);

		if (WRITE(cfd, data_out, data_out_len) <= 0)
			log(LL_DEBUG, "short write on img data");

		free(data_out);
	}
}

void send_jpg_frame(int cfd, source *s, bool get, int quality, const std::vector<filter *> *const filters, resize *const r, const int resize_w, const int resize_h, configuration_t *const cfg, const bool is_view_proxy, const bool repeat)
{
        constexpr char reply_headers[] =
                "HTTP/1.0 200 ok\r\n"
                "cache-control: no-cache\r\n"
                "pragma: no-cache\r\n"
		"server: " NAME " " VERSION "\r\n"
                "expires: thu, 01 dec 1994 16:00:00 gmt\r\n"
                "connection: close\r\n"
                "content-type: image/jpeg\r\n"
                "\r\n";

	if (WRITE(cfd, reply_headers, strlen(reply_headers)) <= 0) {
		log(LL_DEBUG, "short write on response header");
		close(cfd);
		return;
	}

	if (!get)
	{
		close(cfd);
		return;
	}

	set_no_delay(cfd, true);

	bool sc = resize_h != -1 || resize_w != -1;

	uint64_t prev_ts = get_us();
	int w = -1, h = -1;
	uint8_t *work = nullptr;
	size_t work_len = 0;
	if (s -> get_frame((!filters || filters -> empty()) && !sc ? E_JPEG : E_RGB, quality, &prev_ts, &w, &h, &work, &work_len, repeat)) {

		char *data_out = nullptr;
		size_t data_out_len = 0;
		FILE *fh = open_memstream(&data_out, &data_out_len);
		if (!fh)
			error_exit(true, "open_memstream() failed");

		if ((!filters || filters -> empty()) && !sc)
			fwrite(work, work_len, 1, fh);
		else {
			source *cur_s = is_view_proxy ? ((view *)s) -> get_current_source() : s;
			instance_t *inst = find_instance_by_interface(cfg, cur_s);

			apply_filters(inst, cur_s, filters, nullptr, work, prev_ts, w, h);

			if (resize_h != -1 || resize_w != -1) {
				int target_w = resize_w != -1 ? resize_w : w;
				int target_h = resize_h != -1 ? resize_h : h;

				uint8_t *temp = nullptr;
				r -> do_resize(w, h, work, target_w, target_h, &temp);

				write_JPEG_file(s -> get_meta(), fh, target_w, target_h, quality, temp);
				free(temp);
			}
			else {
				write_JPEG_file(s -> get_meta(), fh, w, h, quality, work);
			}
		}

		fclose(fh);

		if (WRITE(cfd, data_out, data_out_len) <= 0)
			log(LL_DEBUG, "short write on img data");

		free(data_out);

		free(work);
	}
}

void send_file(const int cfd, const std::string & path, const char *const name, const bool dl)
{
	std::string complete_path = path + "/" + name;

	std::string type = "text/html";
	if (complete_path.size() >= 3) {
		std::string ext = complete_path.substr(complete_path.size() -3);

		if (ext == "avi")
			type = "video/x-msvideo";
		else if (ext == "mp4")
			type = "video/mp4";
		else if (ext == "png")
			type = "image/png";
		else if (ext == "jpg")
			type = "image/jpeg";
		else if (ext == "css")
			type = "text/css";
		else if (ext == "flv")
			type = "video/x-flv";
		else if (ext == "ico")
			type = "image/x-icon";
		else if (ext == "ebm")
			type = "video/webm";
	}

	log(LL_WARNING, "Sending file %s of type %s", complete_path.c_str(), type.c_str());

	FILE *fh = fopen(complete_path.c_str(), "r");
	if (!fh) {
		log(LL_WARNING, "Cannot access %s: %s", complete_path.c_str(), strerror(errno));

		std::string headers = "HTTP/1.0 404 na\r\nServer: " NAME " " VERSION "\r\n\r\n";

		(void)WRITE(cfd, headers.c_str(), headers.size());

		return;
	}

	std::string name_header = dl ? myformat("Content-Disposition: attachment; filename=\"%s\"\r\n", name) : "";
	std::string headers = "HTTP/1.0 200 OK\r\nServer: " NAME " " VERSION "\r\n" + name_header + "Content-Type: " + type + "\r\n\r\n";
	(void)WRITE(cfd, headers.c_str(), headers.size());

	// FIXME
	while(!feof(fh)) {
		char buffer[4096];

		int rc = fread(buffer, 1, sizeof(buffer), fh);
		if (rc <= 0)
			break;

		if (WRITE(cfd, buffer, rc) <= 0) {
			log(LL_INFO, "Short write");
			break;
		}
	}

	fclose(fh);
}
